/*
 *  RIFFInputStream.java  $Revision: 1.1.1.1 $
 *
 *  $Id: RIFFInputStream.java,v 1.1.1.1 1999/12/20 13:11:59 razi Exp $
 *  $Date: 1999/12/20 13:11:59 $
 *
 *  Copyright (C) 1999  Christian Moench
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Christian Moench
 *  Dept. of Computer Science
 *  Johann Wolfgang Goethe-University
 *  60325 Frankfurt, Germany
 *  Tel.  +49 69 79828358
 *  Fax.  +49 69 79823340
 *
 *  $Log: RIFFInputStream.java,v $
 *  Revision 1.1.1.1  1999/12/20 13:11:59  razi
 *  Lokale Indigo Server auf papageno
 *
 *
 */
import java.io.*;
import java.lang.*;
import java.util.*;


public class RIFFInputStream extends FilterInputStream {

    // The item we are currently parsing
    public ItemInfo  currentItem;

    // The lists we are enrolled in
    public ItemInfo currentList;

    // The current list
    public Stack    upperLists;


    protected void readBytes(byte[] buffer) throws IOException {
	if (in.read(buffer) != buffer.length) {
	    throw new IOException("Short read");
	}
    }


    protected String readForm() throws IOException {
	byte[] buffer = new byte[4];

	readBytes(buffer);
	return new String(buffer);
    }


    protected long readLong() throws IOException {
	byte[] buffer = new byte[4];
	long   result = 0;

	readBytes(buffer);
	for (int i = 0; i < 4; i++) {
	    result = result * 256;
	    if (buffer[3 - i] < 0) {
		result = result + 256 + buffer[3 - i];
	    } else {
		result = result + buffer[3 - i];
	    }
	}
	return result;
    }


    protected void readRIFFHeader() throws IOException {
	long     size;
	String   form;

	form = readForm();
	if (form.equals("RIFF")) {
	    // This is a RIFF file
	    size = readLong();
	    form = readForm();
	    currentItem = new ItemInfo(true, form, size, 0);
	} else {
	    throw new IOException("Not a RIFF tag: " + currentItem.itemForm);
	}
    }


    protected void readItemHeader() throws IOException {
	long     bytesRead;
	long     size;
	String   form;

	form = readForm();
	if (form.equals("LIST")) {
	    // This is a list item
	    size = readLong();
	    form = readForm();
	    currentItem = new ItemInfo(true, form, size, 4);
	    bytesRead = 12;
	} else {
	    // This is a flat item
	    size = readLong();
	    currentItem = new ItemInfo(false, form, size, 0);
	    bytesRead = 8;
	}
	if (inList()) {
	    currentList.itemConsumed = currentList.itemConsumed + bytesRead;
	}
    }


    public RIFFInputStream(InputStream in) throws IOException {
	super(in);

	currentItem = null;
	currentList = null;
	upperLists  = new Stack();
	readRIFFHeader();
    }


    /*
     * Check if we are currently in a list
     */
    public boolean inList() {
	return (currentList != null);
    }


    /*
     * Check if there are more items on the current level, if we are
     * not in a list, no more items are available.
     */
    public boolean hasMoreItems() throws IOException {

	if (inList() == false) {
	    throw new IOException("Not currently in a list");
	}
	if (currentList.itemConsumed < currentList.itemSize) {
	    return true;
	}
	return false;
    }


    /*
     * Descend into the current item, which should be a list item.
     */
    public void enterList() throws IOException {

	/*
	 * Check if the current item is realy a list
	 */
	if (currentItem.isList) {
	    
	    /*
	     * If we are already in a list, push this one.
	     */
	    if (currentList != null) {
		upperLists.push(currentList);
	    }

	    /*
	     * The active item becomes the current list. If the list has
	     * an item, read its header.
	     */
	    currentList = currentItem;

	} else {
	    throw new IOException("Form " + currentItem.itemForm
				  + " is not a list");
	}
    }


    /*
     * Ascend from the currently active list. We should be in a list.
     */
    public void exitList() throws IOException {

	/*
	 * See if we are in a list.
	 */
	if (currentList != null) {
	    long dataSize;

	    // Skip remaining bytes if necessary
	    if (currentList.itemConsumed < currentList.itemSize) {
		super.skip(currentList.itemSize - currentList.itemConsumed);
		currentList.itemConsumed = currentList.itemSize;
	    }

	    // Calculate the length the list uses in enclosing structures.
	    dataSize = currentList.itemSize;

	    /*
	     * If there are lists on the stack, the current list
	     * becomes the list on top of the stack.
	     */
	    if (upperLists.empty() == false) {
		currentList = (ItemInfo)upperLists.pop();
		currentList.itemConsumed =
		    currentList.itemConsumed + dataSize; 
	    } else {
		currentList = null;
	    }
	} else {
	    throw new IOException("Exit while not in list");
	}
    }


    /*
     * Skip the current item and go to the next item, if no items are
     * left, the current item is set to null.
     */
    public void skipItem() throws IOException {

	// Skip the current item in the input file
	super.skip(currentItem.itemSize - currentItem.itemConsumed);
	currentItem.itemConsumed = currentItem.itemSize;

	/*
	 * If we are in a list, set the consumption values and read
	 * the next item.
	 */
	if (inList()) {

	    if (currentItem.isList) {
		currentList.itemConsumed = currentList.itemConsumed 
		    + currentItem.itemSize;
	    } else {
		currentList.itemConsumed = currentList.itemConsumed 
		    + currentItem.itemSize;
	    }
	}
    }


    /*
     * Read bytes from the stream, this method overwrites read in
     * FilterInputStream.
     */
    public int read() throws IOException {
	int result;

	// Read is not allowed in lists.
	if (currentItem.isList) {
	    throw new IOException("read() in list");
	}

	// Prevent reading beyond the borders of the current chunk.
	if (currentItem.itemConsumed < currentItem.itemSize) {
	    result = in.read();
	    if (result > -1) {
		// Set consumed byte counters.
		currentItem.itemConsumed++;
		if (inList()) {
		    currentList.itemConsumed++;
		}
	    }
	    return result;
	}
	return -1;
    }


    /*
     * TODO: Provide efficient implementation for reading a number of
     * bytes at once. The superclass implementation just makes
     * successive calls to read.
     */
    public int read(byte b[], int off, int len) throws IOException {

	// Read is not allowed in lists.
	if (currentItem.isList) {
	    throw new IOException("read() in list");
	}

	// Prevent reading from an offset behind the chunk
	if (currentItem.itemConsumed + off >= currentItem.itemSize) {
	    return -1;
	}

	// Skip offset bytes and add to consumed bytes
	currentItem.itemConsumed = currentItem.itemConsumed + off;
	in.skip(off);

	// Calculate the maximum allowed length readable from chunk
	if (currentItem.itemConsumed + len >= currentItem.itemSize) {
	    len = (int)(currentItem.itemSize - currentItem.itemConsumed);
	}

	// Check read length against buffer length
	if (len > b.length) {
	    len = b.length;
	}

	// Read remaining bytes with an offset of zero.
	int result = in.read(b, 0, len);
	if (result >  -1) {
	    // Set consumed byte counters.
	    currentItem.itemConsumed = currentItem.itemConsumed + result;
	    if (inList()) {
		currentList.itemConsumed = currentList.itemConsumed + result;
	    }
	}
	return result;
    }


    public static void showChunk(String indent,
				 RIFFInputStream in) throws IOException {
	
	System.out.println(indent
			   + "(Chunk: " + in.currentItem.itemForm
			   + ", size: " + in.currentItem.itemSize);
	while (true) {
	    for (int i = 0; i < 16; i++) {
		int value = in.read();
		if (value == -1) {
		    if (i > 0) {
			System.out.println();
		    }
		    System.out.println(indent + ")");
		    return;
		}
		if (i == 0) {
		    System.out.print(indent+ "    ");
		}
		System.out.print(Integer.toString(value) + " ");
	    }
	    System.out.println();
	}
    }


    public static void showList(String indent,
				RIFFInputStream in) throws IOException {

	while (in.hasMoreItems()) {
	    in.readItemHeader();
	    if (in.currentItem.isList) {
		System.out.println(indent
				   + "(List: " + in.currentItem.itemForm
				   + ", size: " + in.currentItem.itemSize);
		in.enterList();
		showList(indent + "    ", in);
		in.exitList();
		System.out.println(indent + ")");
	    } else {
		showChunk(indent, in);
	    }
	}
    }


    public static void main(String[] args) throws IOException {
	RIFFInputStream riffInput;

	riffInput = new RIFFInputStream(new FileInputStream(args[0]));
	if (riffInput.currentItem.isList) {
	    System.out.println("(List: " + riffInput.currentItem.itemForm
			       + ", size: " + riffInput.currentItem.itemSize);
	    riffInput.enterList();
	    showList("    ", riffInput);
	    riffInput.exitList();
	    System.out.println(")");
	} else {
	    showChunk("", riffInput);
	}
    }


    class ItemInfo extends Object {
	
	public boolean  isList;
	public String   itemForm;
	public long     itemSize;
	public long     itemConsumed;

	public ItemInfo(boolean list, String form, long size, long cons) {
	    
	    isList = list;
	    itemForm = form;
	    itemSize = size;
	    itemConsumed = cons;
	} 
    }
}
